 ;
 ;  benchmark program  : biq_N_arc.asm
 ;
 ;  benchmark suite    : DSP-kernel
 ;
 ;  description        : benchmarking of an iir biquad (N sections)
 ;                       
 ;	The equations of the filter are:
 ;       w(n) =    x(n) - ai1*w(n-1) - ai2*w(n-2)
 ;       y(n) = b0*w(n) + bi1*w(n-1) + bi2*w(n-2)
 ;
 ; This routine implements a second order filter with a minimum of
 ; delay elements.  The difference equation relating the output
 ; to the input is:
 ;
 ;     y(n) = a1 * y(n-1) + a2 * y(n-2) + b0 * x(n) + b1 * x(n-1) + b2 * x(n-2)
 ;
 ;  with z transform:
 ;
 ;                          -1       -2    
 ;       Y(z)     b0 + b(1)z  + b(2)z
 ;     ------- = ----------------------
 ;       X(z)              -1       -2
 ;                1 - a(1)z  - a(2)z
 ;
 ; where:
 ;   x(n)  = biquad input sample at time nT
 ;   y(n)  = output of the biquad at time nT
 ;   a(n)  = filter coefficient n (magnitude less than one)
 ;   b(n)  = filter coefficient n (magnitude less than one)
 ;     T   = sample period
 ;
 ;		              w(n)
 ;   x (n)------(-)---------->-|->---b0----(+)-------> y(n)
 ;               A             |            A
 ;               |           |1/z|          |
 ;               |             | w(n-1)     |
 ;               |             v            |
 ;               |-<--a1-----<-|->---b1--->-|
 ;               |             |            |
 ;               |           |1/z|          |
 ;               |             | w(n-2)     |
 ;               |             v            |
 ;               |-<--a2-----<--->---b2--->-|
 ; 
 ;     The filter states w(n-1) and w(n-2) are stored in X data locations
 ;     The coefficients of each biquad are stored in Y data locations
 ;                                              
 ; For the above filter, the memory map  for  the  filter
 ;  states and the coefficients is shown below.
 ;
 ;          r0
 ;          |
 ;          v
 ;      ---------------------------------------------
 ; X:   |        |        |        |        |       |
 ;      | w(n-2) | w(n-1) | w(n-2) | w(n-1) | ...   |   Filter States
 ;      ---------------------------------------------
 ;
 ;      ---------------------------------------------
 ; Y:   |  a(2)  |  a(1)  |  b(0)  |  b(2)  |  b(1) |
 ;      ---------------------------------------------
 ;          ^
 ;          |
 ;          r4
 ;
 ;              Memory Map for the Biquad Filter
 ;
 ; The address register r0 is initialized pointing to  the  first
 ; filter  state  and  r4  is  initialized pointing to the filter
 ; coefficients.
 ;
 ;  reference code     : see Dr.Bub iir7.asm code (modified)
 ;                           
 ;                       see implementing IIR/FIR Filters
 ;                           with Motorola's DSP56000/56001, 1991
 ; 
 ;  func. verification : from separate computation
 ;
 ;  organization       : Aachen University of Technology - IS2
 ;                     : DSP Tools Group
 ;                     : phone   : +49(241)807887
 ;                     : fax     : +49(241)8888195
 ;                     : e-mail  : zivojnov@ert.rwth-aachen.de
 ;
 ;  author             : Juan Martinez Velarde
 ;
 ;  history            : 28-3-1994 assembler reference code creation
 ;                       28-3-1994 assembler reference code profiling
 ;

	section biq_one_arc
	opt so,nomd,cc

N	equ	4
	
        org     x:0
datin:  dc	2	    ; location in X memory of input sample
temp:	dc	0
states: dc      1,1          ; filter states
        dc      1,1
        dc      1,1
        dc      1,1
        
	org     y:0
datout:	dc	0	   ;location in Y memory of output file 
coeff:  dc      1,1,2,1,1  ;coefficients: a1,a2,b0,b1,b2 1st biquad
        dc      1,1,2,1,1  ;coefficients: a1,a2,b0,b1,b2 2nd biquad
        dc      1,1,2,1,1  ;coefficients: a1,a2,b0,b1,b3 3rd biquad
        dc      1,1,2,1,1  ;coefficients: a1,a2,b0,b1,b4 4th biquad

LC0	dc	'output',$3D,$25,'i',$0A,$0A,$00

	org	p:
	global	Fmain
Fmain
	move	ssh,y:(r6)+
 
        move    #states,r0     	; point to filter states
	move	#datout,r1	; point to data output location
	move	#datin,r2      	; point to data input location
	move	#temp,r3	; copy input sample to temp location
	move	x:(r2),a0
	move	a0,x:(r3)
	move	#temp,r2
        move    #coeff,r4 	; point to filter coefficients (a2,a1,b0,b2,b1)

ASM_APP_0
	START_PROFILING:
ASM_APP_1
	
	do	#N,LOOPE			; for N biquads
	move	x:(r3),a0			; a = biquad system input sample
	asl	a        x:(r0)+,x0  y:(r4)+,y0	; x0 = w(n-2), y0 = a2
	mac     x0,y0,a  x:(r0)-,x1  y:(r4)+,y0 ; a  = a2*w(n-2),x1=w(n-1),y0 = a1
        mac     x1,y0,a  x1,x:(r0)+  y:(r4)+,y0 ; a += a1*w(n-1),store w(n-1) as new w(n-2),y0=b0
	asr	a
	move	a0,x:(r0)+
	move	a0,y1
	mpy	y0,y1,a		y:(r4)+,y0
        mac     x0,y0,a         y:(r4)+,y0      ; a += b2*w(n-2),y0=b1
	mac	x1,y0,a	 			; a += b1*w(n-1)
	asr 	a
	move	a0,x:(r3)			; store biquad system output temporaly

LOOPE:
        move    a0,y:(r1)       ; iir filter output sample	

ASM_APP_2
	END_PROFILING:
ASM_APP_3

	clr	a
	move	y:(r1),a0
	move	#LC0,r3
	move	a0,y:(r6)+
	move	r3,y:(r6)+
	jsr	Fprintf
	move 	(r6)-
	move	(r6)-


END:    move	y:-(r6),ssh
	rts

	endsec

