{ 
  Cascaded Biquad IIR Filter Subroutine                     BIQ_ONE_ARC.DSP

   benchmark program  : biquad_one_section.c
 
   benchmark suite    : DSP-kernel
 
   description        : benchmarking of an iir biquad (one section)
                        
       This routine optimizes the one section biquad iir filtering.

 	The equations of the filter are:
        w(n) =    x(n) - ai1*w(n-1) - ai2*w(n-2)
        y(n) = b0*w(n) + bi1*w(n-1) + bi2*w(n-2)
 
 		             w(n)
    x (n)------(-)---------->-|->---b0----(+)-------> y(n)
                A             |            A
                |           |1/z|          |
                |             | w(n-1)     |
                |             v            |
                |-<--a1-----<-|->---b1--->-|
                |             |            |
                |           |1/z|          |
                |             | w(n-2)     |
                |             v            |
                |-<--a2-----<--->---b2--->-|
  

   reference code     :  analog devices' dsp laboratory, chapter 7
                         listing 7-3 iir cascade form subroutine
 
   func. verification : from separate computation
 
   organization       : Aachen University of Technology - IS2
                      : DSP Tools Group
                      : phone   : +49(241)807887
                      : fax     : +49(241)8888195
                      : e-mail  : zivojnov@ert.rwth-aachen.de
 
   author             : Juan Martinez Velarde
 
   history            : 22-3-1994 creation and profiling
 

    Calling Parameters
         I0 --> delay line buffer for x(n-2), x(n-1), y(n-2), y(n-1)
         L0 = 0
         I1 --> list of scale factors for each biquad section
         L1 = number of biquad sections, but 0 for single biquad
	 I4 --> scaled coefficients b2,b1,b0,a2,a1 for a single biquad
         L4 = 2.5 * filter order --or-- 5 * number of biquad sections
	 L4 = 5 (in the case of a single biquad)
         M0,M4 = 1
         M1 = -3
         M2 = 0   (in the case of a single biquad)
         M3 = (1 - length of delay line buffer)
	 I6 --> input sample address   (X value)
	 I7 --> output sample address  (Y value)
	 M7 = L7 = 0 

    Return Values
         SR1 = output sample
         I0 --> inside delay line buffer
         I1 --> top of scale factor list
         I4 --> top of coefficients
	 I7 --> output sample

    Altered Registers
	 I0, M0, L0, I1, M1, L1, M2, M3, I4, M4, L4, I6, M6, L6, I7, M7, L7
         MX0, MX1, MY0, MR, SE, SR

    Computation Time
	12 instruction cycles

    All coefficients and data values are assumed to be in 1.15 format.
}

.MODULE biquad_sub;


.VAR/DM/ABS=0x3400 delayline[4] ; {this is scratchpad memory}
.VAR/DM/ABS=0x3404 scalelist[1] ; {initialize with the scale factor for each biquad}
.VAR/DM/ABS=0x3405 X ; { input sample } 
.VAR/DM/ABS=0x3406 Y ; { output sample }
.VAR/PM/ABS=0x20 coefflist[5];{initialize with 5 filter coefficients for biquad}

.INIT delayline:	1,1,1,1 ; 
.INIT scalelist:	0x01 ; 
.INIT coefflist:	0x200000, 0x400000, 0x200000, 0xF5C800, 0x0ED000  ; 
.INIT X:		0x1 ; 
.INIT Y:		0x0 ; 

.ENTRY  biquad;
	
biquad: 
	{ initialize registers }

	I0=^delayline ; M0 = 1  ; L0=0 ; 
	I1=^scalelist ; M1 = -3 ; L1=1 ; 
	I4=^coefflist ; M4 = 1  ; L4=5 ; 
	M2 = 0; M3 = -3 ; 
	I6 =^X ; M6 = 1 ; L6 = 0 ; 
	I7 =^Y ; M7 = 1 ; L7 = 0 ; 

START_PROFILING:

	SR1=DM(I6,M6) ; 		                 {get x(n)}
        SE=DM(I1,M2);
	MX0=DM(I0,M0), MY0=PM(I4,M4);                    {get x(n-2), B2}
        MR=MX0*MY0(SS), MX1=DM(I0,M0), MY0=PM(I4,M4);    {get x(n-1), B1}
        MR=MR+MX1*MY0(SS), MY0=PM(I4,M4);                {get B0}
        MR=MR+SR1*MY0(SS), MX0=DM(I0,M0); MY0=PM(I4,M4); {get y(n-2), A2}
        MR=MR+MX0*MY0(SS), MX0=DM(I0,M1), MY0=PM(I4,M4); {get y(n-1), A1}

        DM(I0,M0)=MX1, MR=MR+MX0*MY0(RND); {store x(n-1) as new x(n-2)}
        DM(I0,M0)=SR1, SR=ASHIFT MR1 (HI); {store x(n) as new x(n-1),SR1=output}

        DM(I0,M0)=MX0;           {store y(n-1) as new y(n-2)} 
	DM(I0,M3)=SR1 ;          {store y(n) as new y(n-1)}
 	DM(I7,M6)=SR1 ;	         {store y(n) at output address} 

END_PROFILING:

        RTS;

.ENDMOD;
