;   benchmark program:   lms_arc.asm
;   
;   benchmark suite:     DSP-kernel
;   
;   description:         lms - adaptive fir filter benchmarking
; 
; 
;                x(n-0) |----| x(n-1)|----| ......|----| x(n-N+1)
;     x(n) -------------| 1/z|-------| 1/z|-------| 1/z|----
;                   |   |----|   |   |----|   |   |----|   |
;                   h0           h1          ...           hN-1
;                     \________  |____   _____|  _________/
;                              \______(+)_______/
;                                      |
;                                      |--------->y(n)
;                                      v
;           d(n) -------------------->(+)-------->e(n)
;                                    -
; 
; 
; Notation and symbols:
;  x(n)  - Input sample at time n.
;  d(n)  - Desired signal at time n.
;  y(n)  - FIR filter output at time n.
;  H(n)  - Filter coefficient vector at time n.  
;          H={h0,h1,...,hN-1}
;  X(n)  - Filter state variable vector at time N.
;          X={x(0),x(n-1),...,x(n-N+1)}
;  delta - Adaptation gain.
;  N     - Number of coefficient taps in the filter.
; 
; 
;      True LMS Algorithm   
;      ------------------   
;      Get input sample     
;      Save input sample    
;      Do FIR               
;      Get d(n), find e(n)  
;      Update coefficients 
;      Output f(n)         
;      Shift vector X      
; 
; System equations:
;  e(n)=d(n)-H(n)X(n)     (FIR filter and error)
;  H(n+1)=H(n)+uX(n)e(n)  (Coefficient update)
; 
; References:
;  "Adaptive Digital Filters and Signal Analysis", Maurice G. Bellanger
;      Marcel Dekker, Inc. New York and Basel
; 
;  "The DLMS Algorithm Suitable for the Pipelined Realization of Adaptive
;      Filters", Proc. IEEE ASSP Workshop, Academia Sinica, Beijing, 1986
; 
; Note: The sections of code shown describe how to initialize all
;      registers, filter an input sample and do the coefficient update.
;      The instructions relating to the filtering and coefficient
;      update are shown as part of the benchmark.  Instructions executed
;      only once (for initialization) are not included in the benchmark.
; 
;            Implementation of the true LMS on the DSP56000
; 
; 
; Memory map:
; 
;    Initial X                              H
;  x(n-0) x(n-1) x(n-2) ... x(n-N+1)       h0   h1   h2  ... hN-1
;   |                                      |
;   r0                                     r4
;                                          r5
;   
;   reference code:      target assembly
;   
;   f. verification:     
;   
;    organization:        Aachen University of Technology - IS2 
;                         DSP Tools Group
;                         phone:  +49(241)807887 
;                         fax:    +49(241)8888195
;                         e-mail: zivojnov@ert.rwth-aachen.de 
;  
;   author:              Juan Martinez Velarde
;   
;   history:             30-3-94 Assembler Code creation (Martinez Velarde)
;                        30-3-94 Assembler Code profiling (Martinez Velarde) 
;  
	section lms_arc
	opt so,nomd,cc
	
N	equ	16

	org	x:$0 	; samples in x memory
F___x0  dc 1,0,0,0,0,0,0,0,0,0,0,0,0,0,0

input:  dc 1

	org	y:$0	; coefficients in y memory
F___h1:	dc 1
	dc 1
	dc 1
	dc 1	
	dc 1
	dc 1
	dc 1
	dc 1	
	dc 1
	dc 1
	dc 1
	dc 1	
	dc 1
	dc 1
	dc 1
	dc 1

output:  dc 0
desired: dc 1000
delta:   dc 1

LC0	dc	'output',$3D,$25,'i',$0A,$0A,$00
LC1	dc	'first input',$3D,$25,'i',$0A,$0A,$00
LC2	dc	'desired',$3D,$25,'i',$0A,$0A,$00
LC3	dc	'coeff',$3D,$25,'i',$0A,$0A,$00
LC4	dc	'filter input',$3D,$25,'i',$0A,$0A,$00
LC5	dc	'y1',$3D,$25,'i',$0A,$0A,$00
LC6	dc	'states',$3D,$25,'i',$0A,$0A,$00
LC7	dc	'delta',$3D,$25,'i',$0A,$0A,$00
LC8	dc	'y0',$3D,$25,'i',$0A,$0A,$00

printinputsample	macro
printinputsample	ident	1,0

	move	y:-(r6),m5
	move	y:-(r6),m4
	move	y:-(r6),m0

	move	#LC4,r1
	move 	#F___x0,r3	; start of X states
	move	x:(r3),a0
	move	a0,y:(r6)+
	move	r1,y:(r6)+
	jsr	Fprintf
	move 	(r6)-
	move	(r6)-

	move	m0,y:(r6)+
	move	m4,y:(r6)+
	move	m5,y:(r6)+

	move	#N-1,m0		; mod N
	move	m0,m4		; mod N
	move	m0,m5		; mod N

	endm

printy0		macro
printy0		ident	1,0

	move	y:-(r6),m5
	move	y:-(r6),m4
	move	y:-(r6),m0

	move	#LC8,r1
	move	y0,a0
	move	y1,y:(r6)+
	move	y0,y:(r6)+
	move	a0,y:(r6)+
	move	r1,y:(r6)+
	jsr	Fprintf
	move 	(r6)-
	move	(r6)-
	move	y:-(r6),y0
	move	y:-(r6),y1

	move	m0,y:(r6)+
	move	m4,y:(r6)+
	move	m5,y:(r6)+

	move	#N-1,m0		; mod N
	move	m0,m4		; mod N
	move	m0,m5		; mod N

	endm


printoutputsample	macro
printoutputsample	ident	1,0

	move	y:-(r6),m5
	move	y:-(r6),m4
	move	y:-(r6),m0

	move	#LC0,r3
	move	y:output,a0
	move	a0,y:(r6)+
	move	r3,y:(r6)+
	jsr	Fprintf
	move 	(r6)-
	move	(r6)-

	move	m0,y:(r6)+
	move	m4,y:(r6)+
	move	m5,y:(r6)+

	move	#N-1,m0		; mod N
	move	m0,m4		; mod N
	move	m0,m5		; mod N

	endm

printdelta	macro
printdelta	ident	1,0

	move	y:-(r6),m5
	move	y:-(r6),m4
	move	y:-(r6),m0

	move	#LC7,r3
	move	y1,a0
	move	y0,y:(r6)+
	move	y1,y:(r6)+
	move	a0,y:(r6)+
	move	r3,y:(r6)+
	jsr	Fprintf
	move 	(r6)-
	move	(r6)-
	move	y:-(r6),y1
	move	y:-(r6),y0

	move	m0,y:(r6)+
	move	m4,y:(r6)+
	move	m5,y:(r6)+

	move	#N-1,m0		; mod N
	move	m0,m4		; mod N
	move	m0,m5		; mod N

	endm



printcoeffs	macro
printcoeffs	ident 	1,0

	move	y:-(r6),m5
	move	y:-(r6),m4
	move	y:-(r6),m0

	move	#F___h1,r3
	do	#N,LEND1
	clr	a
	move	y:(r3)+,a0
	move	#LC3,r1
	move	a0,y:(r6)+
	move	r1,y:(r6)+
	jsr	Fprintf
	move 	(r6)-
	move	(r6)-
LEND1:

	move	m0,y:(r6)+
	move	m4,y:(r6)+
	move	m5,y:(r6)+

	move	#N-1,m0		; mod N
	move	m0,m4		; mod N
	move	m0,m5		; mod N

	endm


printstates	macro
printstates	ident 	1,0

	move	y:-(r6),m5
	move	y:-(r6),m4
	move	y:-(r6),m0

	move 	#F___x0,r3	; start of X states
	do	#N,LEND2
	move	#LC6,r1
	clr	a
	move	x:(r3)+,a0
	move	a0,y:(r6)+
	move	r1,y:(r6)+
	jsr	Fprintf
	move 	(r6)-
	move	(r6)-
LEND2:
	move	m0,y:(r6)+
	move	m4,y:(r6)+
	move	m5,y:(r6)+

	move	#N-1,m0		; mod N
	move	m0,m4		; mod N
	move	m0,m5		; mod N

	endm

	org	p:$100

	global	Fmain
Fmain
	move	ssh,y:(r6)+
	move	m0,y:(r6)+
	move	m4,y:(r6)+
	move	m5,y:(r6)+

	move 	#F___x0,r0	; start of X states
	move	#F___h1,r4	; start of coefficients
	move	r4,r5		; start of coefficients
	move	#N-1,m0		; mod N-1
	move	#-2,n0		; adjust for filtering
	move	m0,m4		; mod N-1
	move	m0,m5		; mod N-1
	move	x:input,b0	; get input sample in b0

	move	#0,x1

ASM_APP_0
	START_PROFILING:
ASM_APP_1

smploop

;
; FIR filtering
;
	clr	a	
	move	y:(r4)+,y0 			; get h0
	move	x:(r0)+,x0			; get x0
    	rep  	#N-1                       	; do fir
 	mac 	x0,y0,a	x:(r0)+,x0 y:(r4)+,y0  	; do taps
	mac	x0,y0,a				; last tap
	asr	a	b0,x:(r0)+		; save input sample
;
; get desired value, subtract fir output, multiply by delta, 
; put the result in y1. 
;
	move	a0,y:output	; store fir output signal
	move	y:delta,y1	; get delta value
	move	y:desired,y0  	; get desired output

;	printoutputsample
;	printdelta
;	printy0

	neg	a		; - fir output
	add	y0,a		; error = desired - fir output
	move	a,a0
	move	a,y0		; store e in x0
	mpy 	y0,y1,a	 	; a = e * delta
	asr	a
	move	a0,y1		; e*delta stored in y1

;	printoutputsample
;	printdelta
;	printy0

;
; Coefficient Update
;
	move	x:(r0)+,x1	    y:(r4)+,a	; get x(0), h0
	do 	#N,_coefupdate			; do it N times
	mac	x1,y1,a x:(r0)+,x1  y:(r4)+,y0 	; h += delta*e*x
	asr 	a
	move	a0,y:(r5)+			; save new h
	tfr	y0,a				; y0 = a

_coefupdate

;
; Update pointers
;
	move	x:(r0)-,x1 	y:(r4)-,y0		; update r0,r4

ASM_APP_2
	END_PROFILING:
ASM_APP_3

; 	printinputsample
;	printoutputsample
	printstates
	printcoeffs

	jmp	smploop

	move	y:-(r6),m5
	move	y:-(r6),m4
	move	y:-(r6),m0

	move	y:-(r6),ssh

	rts
	endsec

